HonorCapItAll = LibStub('AceAddon-3.0'):NewAddon('HonorCapItAll', 'AceEvent-3.0')
local L = LibStub('AceLocale-3.0'):GetLocale('HonorCapItAll', true)

--   data consists of the relevent currency and quest data we need to calculate
--   honor from them.
--
--   Currency entries consists of a table with the following keys:
--   id = The id of the currency item or Honor or Arena
--   entry_type = 'currency'
--   count =  The current count the character has for that currency
--   honor = The amount of honor the item represents.
--   from = table of the currencies and the purchase cost in them to buy this entry
--
--   Quest entries consists of a table with the following keys:
--   id = The id of the quest
--   entry_type = 'quest'
--   min_level = minimum level you can turn in the queset
--   max_level = maximum level yu can turn in the quest
--   faction = Alliance or Horde
--   items = table of the itemids of things required to complete the quest and the count
--   honor = The amount of honor the quest rewards.
--
--   Note that not every quest needs to go in here, only quests which do not show in
--   your quest log, such as the BG mark repeatables.
--
--   The order of this table matters in two ways.
--   1) It dettermines the order in which everything is processed.  So that if in the
--      future Blizzards adds a way to receive honor from the same source but to different
--      degrees you can prioritize methods by best honor return.  E.G. the old 2k for all 4
--      battleground tokens vs 300 honor for 1.
--   2) It controls the order in which the items are displayed in the tooltip.  

local data = {
	{ 
		id = 20558, -- Warsong Gulch Mark of Honor
		entry_type = 'currency',
	},
	{
		id = 20559, -- Arathi basin Mark of Honor
		entry_type = 'currency',
	},
	{
		id = 20560, -- Alterac Valey Mark of Honor
		entry_type = 'currency',
	},
	{
		id = 29024, -- Eye of the Storm Mark of Honor
		entry_type = 'currency',
	}, 
	{
		id = 42425, -- Strand of the Ancients Mark of Honor
		entry_type = 'currency',
	}, 
	{
		id = 47395, -- Isle of Conquest Mark of Honor
		entry_type = 'currency',
	},
	{ 
		id=8371, -- Concerted Efforts
		entry_type = 'quest',
		min_level=61,
		max_level=70,
		faction='Alliance',
		items = { [20558]=1, [20559]=1, [20560]=1, [29024]=1 },
		honor=314
	},
	{ 
		id=13478, -- Concerted Efforts
		entry_type = 'quest',
		min_level=71,
		max_level=80,
		faction='Alliance',
		items = { [20558]=1, [20559]=1, [20560]=1, [29024]=1, [42425]=1, [47395]=1 },
		honor=1489
	},
	{ 
		id=8367, -- For Great Honor
		entry_type = 'quest',
		min_level=61,
		max_level=70,
		faction='Horde',
		items = { [20558]=1, [20559]=1, [20560]=1, [29024]=1 },
		honor=314
	}, 
	{
		id=13476, -- For Great Honor
		entry_type = 'quest',
		min_level=71,
		max_level=80,
		items = { [20558]=1, [20559]=1, [20560]=1, [29024]=1, [42425]=1, [47395]=1 },
		faction='Horde',
		honor=1489
	},
	{
		id = 43228, -- Stone Keeper's Shard
		entry_type = 'currency',
		},
	{
		id = 43589, -- Wintergrsp Mark of Honor
		entry_type = 'currency',
	},
	{
		id = 44115, -- Wintergrasp Commendation
		entry_type = 'currency',
		honor = 2000,
		from = { [43228] = 30, [43589] = 9 }
	},
	{ 
		id = 'questlog', -- special entry type to trigger the quest log processing.
	  entry_type = 'questlog',  -- This entry must be after the arena points but before
		                          -- anything you buy with them.
	},
	{
		id = 45706, -- Commendation of Bravery
		entry_type = 'currency',
		honor = 2000,
		from = { ['Arena'] = 100 }
	},
}

-- Generator to return the next currency in the data table
local function next_currency(data, i)
	i = next(data, i)
	if i == nil then
		return
	end
	local entry = data[i]
	if entry.entry_type == 'currency' then
		return i, entry
	end
	return next_currency(data, i)
end

-- Generator to return the next quest in the data table
local function next_quest(data, i)
	i = next(data, i)
	if i == nil then
		return
	end
	local entry = data[i]
	if entry.entry_type == 'quest' then
		return i, entry
	end
	return next_quest(data, i)
end

-- Metatable to ensure that a table returns 0 for any missing entries
local zero_mt = {
	__index = function(t, k)
		if k == nil then return end
		t[k] = 0
		return 0
	end,
}

local DATABASE_DEFAULTS = {
	profile = {
		include_questlog = true,
		max = {
			[45706] = 5,
			['**'] = 0,
		},
		set_aside = {
			['**'] = 0,
		},
	}
}

-- currency_counts tracks the amount of any given currency we have.
local currency_counts = { }
setmetatable(currency_counts, zero_mt)

-- quest_honor tracks the honor given by any quests in your quest log
-- quest_arena tracks the arena points ggiven by any quests in your quest log
local quest_honor = { }
setmetatable(quest_honor, zero_mt)
local quest_arena = { }
setmetatable(quest_arena, zero_mt)

-- Keep track of localized quest names. 
local quest_titles = {
	-- Include some English defaults in case the tooltip scanning fails
	-- for some reason.
	[8371]  = 'Concerted Efforts',
	[13478] = 'Concerted Efforts',
	[8367]  = 'For Great Honor',
	[13476] = 'For Great Honor',
}

-- Iterate the currnecies and update our counts for them.
local function update_currency_counts()
	for _, entry in next_currency, data do
		local id = entry.id
		currency_counts[id] = GetItemCount(id, true --[[include bank]])
	end

	-- Special casing for Honor and Arena 
	currency_counts['Honor'] = GetHonorCurrency()
	currency_counts['Arena'] = GetArenaCurrency()
end

-- Update the quest_honor and quest_arena tables
local function update_quest_counts()
	wipe(quest_honor)
	wipe(quest_arena)

	if not HonorCapItAll.db.profile.include_questlog then return end

	local saved_i = GetQuestLogSelection()
	for i=1,GetNumQuestLogEntries() do
		local title,_,_,_,_,isCollapsed,isComplete,_,quest_id = GetQuestLogTitle(i)
		if isCollapsed then
			local count = GetNumQuestLogEntries()
			ExpandQuestHeader(i)
			count = GetNumQuestLogEntries() - count
			for j = i +  1, i + count do
				local title,_,_,_,_,_,isComplete,_,quest_id = GetQuestLogTitle(j)
				if isComplete and isComplete > 0 then
					SelectQuestLogEntry(j)
					local honor = GetQuestLogRewardHonor()
					local arena = GetQuestLogRewardArenaPoints()
					if honor > 0 or arena > 0 then
						quest_honor[quest_id] = honor
						quest_arena[quest_id] = arena
						quest_titles[quest_id] = title
					end
				end
			end
			CollapseQuestHeader(i)
		elseif isComplete and isComplete > 0 then
			SelectQuestLogEntry(i)
			local honor = GetQuestLogRewardHonor()
			local arena = GetQuestLogRewardArenaPoints()
			if honor > 0 or arena > 0 then
				quest_honor[quest_id] = honor
				quest_arena[quest_id] = arena
				quest_titles[quest_id] = title
			end
		end
	end
	SelectQuestLogEntry(saved_i)
end

-- Table we build to keep track of the honor calculations.  This is what we build our
-- tooltip off of.  
--
-- It's indexed with each value set to a table.  Each table has the following keys:
-- src_id = The source item or quest.
-- src_count = The number of the item we have or the number of turnins we can do.
-- src_type = 'quest' or 'currency'
-- dst_id = The item the source converted into or nil if converted to honor
-- dst_count = The number of items the conversion created or nil if honor
-- dst_type = 'quest' or 'currency'
-- honor = The amount of honor this produced.
-- arena = The amount of arena points this produced.
local honor_cache = { }

-- Helper function for update_honor_cache().  It sets the specified position in the honor_cache to the passed
-- values.
local function set_honor_cache_entry(pos, src_id, src_count, src_type, dst_id, dst_count, dst_type, honor, arena)
	local entry = honor_cache[pos]
	if not entry then
		entry = { }
		honor_cache[pos] = entry
	end

	entry.src_id = src_id
	entry.src_count = src_count
	entry.src_type = src_type
	entry.dst_id = dst_id
	entry.dst_count = dst_count
	entry.dst_type = dst_type
	entry.honor = honor
	entry.arena = arena
end

local player_faction = UnitFactionGroup('player')
local function update_honor_cache()
	local position = 1
	local spent = { }
	setmetatable(spent, zero_mt)
	local extra = { }
	setmetatable(extra, zero_mt)

	-- Copy the setasides to the spent table, since they're basically the same
	for k,v in pairs(HonorCapItAll.db.profile.set_aside) do
		spent[k] = v
	end

	local player_level = UnitLevel('player')

	for _, entry in ipairs(data) do
		local entry_type = entry.entry_type
		if entry_type == 'currency' then
			local id, honor, arena, from = entry.id, entry.honor, entry.arena, entry.from
			local count = currency_counts[id] - spent[id] + extra[id]
			if count < 0 then count = 0 end
			local max = HonorCapItAll.db.profile.max[id]

			if honor then
				-- Handles the conversion of the a token to honor
				set_honor_cache_entry(position, id, count, entry_type, nil, nil, nil, count * honor, 0)
				spent[id] = spent[id] + count
				position = position + 1
			end

			if from then
				-- Figures out how many of a given token you can purchase
				for from_id, cost in pairs(from) do
					local from_count = currency_counts[from_id] - spent[from_id] + extra[from_id]
					if from_count < 0 then from_count = 0 end
					local bought = math.floor(from_count/cost)
					if max and max > 0 then
						bought = math.min(bought,max - count)
					end
					set_honor_cache_entry(position, from_id, from_count, entry_type, id, bought, entry_type, honor * bought, 0)
					position = position + 1
					spent[from_id] = spent[from_id] + count
					count = count + bought
				end
			end
		elseif entry_type == 'quest' then
			if player_faction == entry.faction and player_level >= entry.min_level and player_level <= entry.max_level then
				local items = entry.items
				if items then
					local quest_id = entry.id
					local turnins_possible 
					for item_id, required in pairs(items) do
						local count = math.floor((currency_counts[item_id] - spent[item_id] + extra[item_id]) / required)
						if not turnins_possible or count < turnins_possible then
							turnins_possible = count
						end
					end
					if turnins_possible then
						set_honor_cache_entry(position, quest_id, turnins_possible, entry_type, nil, nil, nil, entry.honor * turnins_possible, 0)
						for item_id, required in pairs(items) do
							spent[item_id] = spent[item_id] + (required * turnins_possible)
						end
						position = position + 1
					end
				end
			end
		elseif entry_type == 'questlog' then
			for quest_id, honor in pairs(quest_honor) do
				local arena = quest_arena[quest_id]
				if honor > 0 or arena > 0 then
					set_honor_cache_entry(position, quest_id, 1, 'quest', nil, nil, nil, honor, arena)
					position = position + 1
				end

				if arena > 0 then
					extra['Arena'] = extra['Arena'] + arena
				end
			end
		end	
	end

	-- Wipe out any extra entries beyond what we used.
	for i = position + 1, #honor_cache do
		set_honor_cache_entry(i, nil, nil, nil, nil, 0, 0)
	end
end

-- Pretty self explanatory, returns the total honor from our calculations
function HonorCapItAll:GetTotalHonor()
	local total = currency_counts['Honor']
	for _,entry in pairs(honor_cache) do
		total = total + entry.honor
	end
	return total
end

-- Build the frame we use to display total honor on the PVPFrame
local PVPFrameHonorCapItAll = CreateFrame('Frame', 'PVPFrameHonorCapItAll', PVPFrameHonor)
PVPFrameHonorCapItAll:SetWidth(75)
PVPFrameHonorCapItAll:SetHeight(20)
PVPFrameHonorCapItAll:EnableMouse(true)
PVPFrameHonorCapItAll.honor = PVPFrameHonorCapItAll:CreateFontString(nil, nil, 'GameFontNormal')
PVPFrameHonorCapItAll.honor:SetAllPoints(PVPFrameHonorCapItAll)

-- Adjust the positioning of the frames to leave room for our text.
local function adjust_pvp_frame()
	PVPFrameHonor:ClearAllPoints()
	PVPFrameHonor:SetPoint('TOPLEFT',PVPFrameBackground,'TOPLEFT',78,-2)
	PVPFrameHonorCapItAll:ClearAllPoints()
	PVPFrameHonorCapItAll:SetPoint('LEFT',PVPFrameHonorPoints,'RIGHT')
	PVPFrameHonorIcon:ClearAllPoints()
	PVPFrameHonorIcon:SetPoint('LEFT',PVPFrameHonorCapItAll,'RIGHT',4,-6)
	PVPFrameHonorCapItAll:Show()
end

-- Set the text on our frame on the PVPFrame.
local function update_pvp_frame()
	PVPFrameHonorCapItAll.honor:SetText(format(' / %7d',HonorCapItAll:GetTotalHonor()))
end

-- Helper function for show_tooltip
local function format_value(value, value_type, value_id)
	if value_type == 'currency' then
		if value_id == 'Honor' then
			return format('%d |TInterface\\TargetingFrame\\UI-PVP-%s:0:0:0:0:64:64:5:37:3:37|t',value,player_faction)
		elseif value_id == 'Arena' then
			return format('%d |TInterface\\PVPFrame\\PVP-ArenaPoints-Icon:0|t',value)
		else
			return format('|T%s:0|t x%d',GetItemIcon(value_id),value)
		end
	elseif value_type == 'quest' then
		if value == 1 then
			return quest_titles[value_id]
		else
			return format('%s x%d',quest_titles[value_id],value)
		end
	end
end

-- Just a wrapper for format_value(honor, 'currency', 'Honor') to save some typing
local function format_honor(honor)
	return format_value(honor, 'currency', 'Honor')
end

-- Just a wrapper for format_value(arena, 'currency', 'Arena') to save some typing
local function format_arena(arena)
	return format_value(arena, 'currency', 'Arena')
end

local function show_tooltip()
	GameTooltip_SetDefaultAnchor(GameTooltip, PVPFrameHonorCapItAll)
	GameTooltip:SetText('HonorCapItAll')
	GameTooltip:AddLine(' ')
	local total_honor = currency_counts['Honor']
	GameTooltip:AddDoubleLine(L['Current Honor'],format_honor(total_honor),1,1,1,1,1,1)
	GameTooltip:AddDoubleLine(L['Current Arena'],format_arena(currency_counts['Arena']),1,1,1,1,1,1)
	GameTooltip:AddLine(' ')
	for _,entry in pairs(honor_cache) do
		local honor, arena = entry.honor, entry.arena
		if honor > 0 or arena > 0 then
			local src_id, src_count, src_type = entry.src_id, entry.src_count, entry.src_type
			local dst_id, dst_count, dst_type = entry.dst_id, entry.dst_count, entry.dst_type
			local left, right

			if src_type then
				left = format_value(src_count,src_type,src_id)
			end
	
			if dst_type then
				left = left .. ' = ' .. format_value(dst_count,dst_type,dst_id)
			end

			left = left .. ' = '

			if arena > 0 and honor > 0 then
				right = format_arena(arena) .. ' ' .. format_honor(honor)
			elseif honor > 0 then
				right = format_honor(honor)
			else -- arena > 0
				right = format_arena(arena)
			end

			total_honor = total_honor + honor
			GameTooltip:AddDoubleLine(left,right,1,1,1,1,1,1)
		end
	end
	
	GameTooltip:AddLine(' ')
	GameTooltip:AddDoubleLine(L['Total Honor'],format_honor(total_honor),1,1,1,1,1,1)
	GameTooltip:Show()
end

local function hide_tooltip()
	GameTooltip:Hide()
end

PVPFrameHonorCapItAll:SetScript('OnEnter',show_tooltip)
PVPFrameHonorCapItAll:SetScript('OnLeave',hide_tooltip)

function update_tooltip()
	if GameTooltip:GetOwner() == PVPFrameHonorCapItAll then
		show_tooltip()
	end
end

local function build_config()
	local AceConfig = LibStub('AceConfig-3.0')
	local AceConfigDialog = LibStub('AceConfigDialog-3.0')

	local max_order = 1
	local max_group = {
		desc = {
			type = 'description',
			name = L['Set the maximum number of this item to buy, including ones that you already own.  This is useful for limiting the Commendation of Bravery at the end of the arena season since it is Unique(5).'],
			order = max_order,
		},
		spacer = {
			type = 'description',
			name = '',
			order = max_order + 1,
		},
	}
	max_order = max_order + 2

	local set_aside_order = 1
	local set_aside_group = {
		desc = {
			type = 'description',
			name = L['Set aside some of a given currency.  Useful if you wish to account for planned spending on items that you might buy with these currencies other than converting them to honor.  You could use it to prevent a currency from buying any tokens, for example setting the arena point set aside to a really high number can be used to prevent your arena points from being calculated to use for honor.'],
			order = set_aside_order,
		},
		spacer = {
			type = 'description',
			name = '',
			order = set_aside_order + 1,
		},
	}
	set_aside_order = set_aside_order + 2

	local function name(info)
		local id = info[#info]
		local icon
		if id == 'Arena' then
			icon = 'Interface\\PVPFrame\\PVP-ArenaPoints-Icon'
		else
			icon = GetItemIcon(id)
		end
		return format(' |T%s:32|t',icon)
	end

	local function desc(info)
		local id = info[#info]
		if id == 'Arena' then
			return ARENA_POINTS
		else
			return GetItemInfo(id)
		end
	end

	local function get(info)
		local table = info[#info-1]
		local id = tonumber(info[#info]) or info[#info]
		return tostring(HonorCapItAll.db.profile[table][id])
	end

	local function set(info, value)
		local table = info[#info-1]
		local id = tonumber(info[#info]) or info[#info]
		HonorCapItAll.db.profile[table][id] = tonumber(value)
		update_honor_cache()
		update_pvp_frame()
		update_tooltip()
	end

	local function validate(info, value)
		if not tonumber(value) then
			return L['Must be a number']
		end
		return true
	end

	local max = {
		name = name,
		desc = desc,
		type = 'input',
		get = get,
		set = set,
		validate = validate,
		width = 'half',
	}
	local set_aside = {
		name = name,
		desc = desc,
		type = 'input',
		get = get,
		set = set,
		validate = validate,
		width = 'half',
	}

	local arena = { }
	for k,v in pairs(set_aside) do
		arena[k] = v
	end
	arena.order = set_aside_order
	set_aside_order = set_aside_order + 1
	set_aside_group['Arena'] = arena
	
	for _,entry in next_currency, data do
		if entry.from then
			local config_entry = { }
			for k,v in pairs(max) do
				config_entry[k] = v
			end
			config_entry.order = max_order
			max_order = max_order + 1
			max_group[tostring(entry.id)] = config_entry
		else
			local config_entry = { }
			for k,v in pairs(set_aside) do
				config_entry[k] = v
			end
			config_entry.order = set_aside_order
			set_aside_order = set_aside_order + 1
			set_aside_group[tostring(entry.id)] = set_aside
		end
	end

	local config = {
		name = 'HonorCapItAll',
		type = 'group',
		args = {
			max = {
				name = L['Maximum'],
				type = 'group',
				inline = true,
				args = max_group,
				order = 1,
			},
			set_aside = {
				name = L['Set Aside'],
				type = 'group',
				inline = true,
				args = set_aside_group,
				order = 2,
			},
			misc = {
				name = L['Miscellaneous'],
				type = 'group',
				inline = true,
				order = 3,
				args = {
					include_questlog = {
						type = 'toggle',
						name = L['Include Completed Quests'],
						desc = L['Include completed quests from your quest log in the results.'],
						order = 1,
						get = function(info)
							return HonorCapItAll.db.profile.include_questlog
						end,
						set = function(info, value)
							HonorCapItAll.db.profile.include_questlog = value
							update_quest_counts()
							update_honor_cache()
							update_pvp_frame()
							update_tooltip()
						end,
						width = 'double',
					},
				},
			},
		},
	}

	local profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(HonorCapItAll.db)

	AceConfig:RegisterOptionsTable('HonorCapItAll',config)
	AceConfigDialog:AddToBlizOptions('HonorCapItAll','HonorCapItAll')
	AceConfig:RegisterOptionsTable('HonorCapItAll.profile', profile)
	AceConfigDialog:AddToBlizOptions('HonorCapItAll.profile', profile.name, 'HonorCapItAll')


	-- no longer need ourself
	build_config = nil
end

function HonorCapItAll:OnInitialize()
	-- Fill quest_titles the quests in our data table.
	local tt = CreateFrame('GameTooltip')
	local tt_left = tt:CreateFontString(nil, nil, 'GameTooltipText')
	local tt_right = tt:CreateFontString(nil, nil, 'GameTooltipText')
	tt:AddFontStrings(tt_left, tt_right)
	tt:SetOwner(UIParent, 'ANCHOR_NONE')
	for _,entry in next_quest, data do
		local id = entry.id
		tt:SetHyperlink('quest:' .. id)
		quest_titles[id] = tt_left:GetText()
	end

	db = LibStub('AceDB-3.0'):New('HonorCapItAllDB', DATABASE_DEFAULTS, 'Default')
	DATABASE_DEFAULTS = nil
	self.db = db

	db.RegisterCallback(self, 'OnProfileChanged')
	db.RegisterCallback(self, 'OnProfileCopied', 'OnProfileChanged')
	db.RegisterCallback(self, 'OnProfileReset', 'OnProfileChanged')

	build_config()
end

function HonorCapItAll:OnEnable()
	-- Events that cause us to update the currency table
	self:RegisterEvent('CURRENCY_DISPLAY_UPDATE')
	self:RegisterEvent('KNOWN_CURRENCY_TYPES_UPDATE','CURRENCY_DISPLAY_UPDATE')
	self:RegisterEvent('BAG_UPDATE','CURRENCY_DISPLAY_UPDATE')
	self:RegisterEvent('PLAYERBANKSLOTS_CHANGED','CURRENCY_DISPLAY_UPDATE')

	-- Events for the quest table
	self:RegisterEvent('QUEST_LOG_UPDATE')
	self:RegisterEvent('UNIT_QUEST_LOG_UPDATE')

	-- Events that simply force a recalculate of honor
	self:RegisterEvent('PLAYER_LEVEL_UP')

	-- Fill our tables
	update_currency_counts()
	update_quest_counts()
	update_honor_cache()
	update_tooltip()

	adjust_pvp_frame()
	update_pvp_frame()
end

function HonorCapItAll:OnProfileChanged()
	update_honor_cache()
	update_pvp_frame()
	update_tooltip()
end

function HonorCapItAll:PLAYER_LEVEL_UP()
	update_honor_cache()
	update_pvp_frame()
	update_tooltip()
end

function HonorCapItAll:QUEST_LOG_UPDATE()
	update_quest_counts()
	update_honor_cache()
	update_pvp_frame()
	update_tooltip()

	-- We only need this the first time to fill in our quest cache
	-- after that UNIT_QUEST_LOG_UPDATE is what we want.
	-- The UNIT event is better but does not fire on login.
	self:UnregisterEvent('QUEST_LOG_UPDATE')
end

function HonorCapItAll:UNIT_QUEST_LOG_UPDATE(event, unit)
	if unit ~= 'player' then return end
	update_quest_counts()
	update_honor_cache()
	update_pvp_frame()
	update_tooltip()
end

function HonorCapItAll:CURRENCY_DISPLAY_UPDATE()
	update_currency_counts()
	update_honor_cache()
	update_pvp_frame()
	update_tooltip()
end



